# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for step_wise."""

import unittest
import pyglove.core as pg
from pyglove.ext.early_stopping import step_wise


def get_stopping_steps(policy,
                       num_trials,
                       num_steps,
                       simulate_process_abort=False):
  """Gets early stopping steps based on policy."""
  trials = [pg.tuning.Trial(i + 1, dna=pg.DNA(i), created_time=0)
            for i in range(num_trials)]
  assert not any(policy.should_stop_early(t) for t in trials)
  stopping_trial_steps = []
  for step in range(1, num_steps + 1):
    if simulate_process_abort and step == num_steps - 1:
      return trials, stopping_trial_steps
    for i in range(num_trials):
      if trials[i].status == 'COMPLETED':
        continue
      trial_id = trials[i].id
      reward = float(1 / trial_id * step)
      trials[i].measurements.append(pg.tuning.Measurement(
          step=step, elapse_secs=float(step),
          reward=reward,
          metrics={'loss': trial_id * step / 10}))
      should_stop = policy.should_stop_early(trials[i])

      # Make another call to should_stop_early to test that a measurement
      # will not be inserted to the histogram twice.
      assert should_stop == policy.should_stop_early(trials[i])

      if should_stop:
        trials[i].status = 'COMPLETED'
        trials[i].infeasible = True
        stopping_trial_steps.append((trial_id, step))
  return trials, stopping_trial_steps


class EarlyStopByValueTest(unittest.TestCase):
  """Tests for early_stop_by_value."""

  def test_bad_gating_rules(self):
    with self.assertRaisesRegex(
        ValueError, 'Invalid definition in `step_values`'):
      _ = step_wise.early_stop_by_value([1])()
    with self.assertRaisesRegex(
        ValueError, 'Invalid definition in `step_values`'):
      _ = step_wise.early_stop_by_value([
          (1, 2, 3)
      ])()
    with self.assertRaisesRegex(
        ValueError, 'Invalid definition in `step_values`'):
      _ = step_wise.early_stop_by_value([
          (1, 'abc')
      ])()

  def test_maximize_the_reward(self):
    policy = step_wise.early_stop_by_value([
        # Gate 1 at step 1, stop when reward < 0.6.
        (1, 0.5),

        # Gate 2 at step 3, stop when reward < 2.5.
        (3, 2.5),
    ])()

    # STEP 1
    # Trial 1, step 1, reward 1.00: should stop=False
    # Trial 2, step 1, reward 0.50: should stop=False
    # Trial 3, step 1, reward 0.33: should stop=True
    # Trial 4, step 1, reward 0.25: should stop=True
    # STEP 2
    # Trial 1, step 2, reward 2.00: should stop=False
    # Trial 2, step 2, reward 1.00: should stop=False
    # STEP 3
    # Trial 1, step 3, reward 3.00: should stop=False
    # Trial 2, step 3, reward 1.50: should stop=True
    # STEP 4
    # Trial 1, step 4, reward 4.00: should stop=False
    # STEP 5
    # Trial 1, step 5, reward 5.00: should stop=False
    self.assertEqual(
        get_stopping_steps(policy, 4, 5)[1],
        [(3, 1), (4, 1), (2, 3)])

  def test_minimize_a_metric(self):
    policy = step_wise.early_stop_by_value([
        # Unorder the gating steps to make sure it still work.
        # Gate 1 at step 3, stop when loss > 0.8.
        (3, 0.8),

        # Gate 2 at step 1, stop when loss > 0.3.
        (1, 0.3),
    ], maximize=False, metric='loss')()

    # STEP 1
    # Trial 1, step 1, loss 0.10: should stop=False
    # Trial 2, step 1, loss 0.20: should stop=False
    # Trial 3, step 1, loss 0.30: should stop=False
    # Trial 4, step 1, loss 0.40: should stop=True
    # STEP 2
    # Trial 1, step 2, loss 0.20: should stop=False
    # Trial 2, step 2, loss 0.40: should stop=False
    # Trial 3, step 2, loss 0.60: should stop=False
    # STEP 3
    # Trial 1, step 3, loss 0.30: should stop=False
    # Trial 2, step 3, loss 0.60: should stop=False
    # Trial 3, step 3, loss 0.90: should stop=True
    # STEP 4
    # Trial 1, step 4, loss 0.40: should stop=False
    # Trial 2, step 4, loss 0.80: should stop=False
    # STEP 5
    # Trial 1, step 5, loss 0.50: should stop=False
    # Trial 2, step 5, loss 1.00: should stop=False
    self.assertEqual(
        get_stopping_steps(policy, 4, 5)[1],
        [(4, 1), (3, 3)])

  def test_callable_metrics(self):
    policy = step_wise.early_stop_by_value([
        # Gate 1 at step 1, stop when loss > 0.3.
        (1, 0.3),

        # Gate 2 at step 3, stop when loss > 0.8.
        (3, 0.8),
    ], maximize=False, metric=lambda m: m.metrics['loss'])()

    self.assertEqual(
        get_stopping_steps(policy, 4, 5)[1],
        [(4, 1), (3, 3)])


class EarlyStopByRankTest(unittest.TestCase):
  """Tests for early_stop_by_rank."""

  def test_bad_gating_rules(self):
    with self.assertRaisesRegex(
        ValueError, 'Invalid definition in `step_ranks`'):
      _ = step_wise.early_stop_by_rank([
          (1, 2)
      ])()
    with self.assertRaisesRegex(
        ValueError, 'Invalid definition in `step_ranks`'):
      _ = step_wise.early_stop_by_rank([
          (1, 2, None)
      ])()
    with self.assertRaisesRegex(
        ValueError, 'Invalid definition in `step_ranks`'):
      _ = step_wise.early_stop_by_rank([
          (0.2, 2, 1)
      ])()
    with self.assertRaisesRegex(
        ValueError, r'Rank must be within range \[0.0, 1.0\]'):
      _ = step_wise.early_stop_by_rank([
          (1, 2.0, 1)
      ])()

  def test_maximize_the_reward(self):
    policy = step_wise.early_stop_by_rank([
        # Gate 1 at step 1, stop when reward < top 50% and len(hist) >= 2.
        (1, 0.5, 2),

        # Gate 2 at step 2, stop when reward < top 100.
        (2, 100, 0),

        # Gate 3 at step 3, stop when reward < top 1.
        (3, 1, 0),
    ])()

    # STEP 1
    # Trial 1, step 1, reward 1.00: should stop=False
    # Trial 2, step 1, reward 0.50: should stop=False
    # Trial 3, step 1, reward 0.33: should stop=True
    # Trial 4, step 1, reward 0.25: should stop=True
    # STEP 2
    # Trial 1, step 2, reward 2.00: should stop=False
    # Trial 2, step 2, reward 1.00: should stop=False
    # STEP 3
    # Trial 1, step 3, reward 3.00: should stop=False
    # Trial 2, step 3, reward 1.50: should stop=True
    # STEP 4
    # Trial 1, step 4, reward 4.00: should stop=False
    # STEP 5
    # Trial 1, step 5, reward 5.00: should stop=False
    self.assertEqual(
        get_stopping_steps(policy, 4, 5)[1],
        [(3, 1), (4, 1), (2, 3)])

  def test_minimize_a_metric(self):
    policy = step_wise.early_stop_by_rank([
        # Gate 1 at step 1, stop when loss rank < top % 80 and len(hist) >= 3.
        (1, 0.8, 3),

        # Gate 2 at step 3, stop when loss rank < top 1.
        (3, 1, 0),
    ], maximize=False, metric='loss')()

    # STEP 1
    # Trial 1, step 1, loss 0.10: should stop=False
    # Trial 2, step 1, loss 0.20: should stop=False
    # Trial 3, step 1, loss 0.30: should stop=False
    # Trial 4, step 1, loss 0.40: should stop=True
    # STEP 2
    # Trial 1, step 2, loss 0.20: should stop=False
    # Trial 2, step 2, loss 0.40: should stop=False
    # Trial 3, step 2, loss 0.60: should stop=False
    # STEP 3
    # Trial 1, step 3, loss 0.30: should stop=False
    # Trial 2, step 3, loss 0.60: should stop=True
    # Trial 3, step 3, loss 0.90: should stop=True
    # STEP 4
    # Trial 1, step 4, loss 0.40: should stop=False
    # STEP 5
    # Trial 1, step 5, loss 0.50: should stop=False
    self.assertEqual(
        get_stopping_steps(policy, 4, 5)[1],
        [(4, 1), (2, 3), (3, 3)])

  def test_callable_metric(self):
    policy = step_wise.early_stop_by_rank([
        # Gate 1 at step 1, stop when loss rank < top % 80 and len(hist) >= 3.
        (1, 0.8, 3),

        # Gate 2 at step 3, stop when loss rank < top 1.
        (3, 1, 0),
    ], maximize=False, metric=lambda m: m.metrics['loss'])()
    self.assertEqual(
        get_stopping_steps(policy, 4, 5)[1],
        [(4, 1), (2, 3), (3, 3)])

  def test_recover(self):
    policy = step_wise.early_stop_by_rank([
        # Gate 1 at step 1, stop when reward < top 50% and len(hist) >= 2.
        (1, 0.5, 2),

        # Gate 2 at step 3, stop when reward < top 1.
        (3, 1, 0),
    ])
    p1 = policy()
    trials = get_stopping_steps(p1, 4, 5, simulate_process_abort=True)[0]
    p2 = policy()
    p2.recover(trials)
    self.assertEqual(p1._gate_history, p2._gate_history)
    self.assertEqual(p1._trial_gate_decision, p2._trial_gate_decision)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import inspect
import time
import unittest
from pyglove.core.coding import errors
from pyglove.core.coding import execution
from pyglove.core.coding import permissions


class EvaluateTest(unittest.TestCase):

  def test_with_context(self):
    with execution.context(x=1, y=0):
      with execution.context(x=2, z=2):
        self.assertEqual(
            execution.evaluate(
                inspect.cleandoc(
                    """
                    p = x + y + z
                    """
                ),
                # Override value from the context.
                global_vars=dict(z=3),
                outputs_intermediate=True,
            ),
            dict(p=2 + 0 + 3, __result__=2 + 0 + 3, __stdout__=''),
        )

  def test_basics(self):
    self.assertEqual(
        execution.evaluate(
            inspect.cleandoc(
                """
                x = 1
                y = x + 1
                print(y)
                z = x + y
                """
            ),
            outputs_intermediate=True,
        ),
        dict(x=1, y=2, z=3, __result__=3, __stdout__='2\n'),
    )
    self.assertEqual(
        execution.evaluate(
            inspect.cleandoc(
                """
                x = 1
                y = x + 1
                print(y)
                z = x + y
                """
            ),
        ),
        3,
    )
    with self.assertRaisesRegex(errors.CodeError, 'ValueError'):
      execution.evaluate(
          inspect.cleandoc(
              """
              def foo():
                raise ValueError("intentional error")
              foo()
              """
          ),
          permission=permissions.CodePermission.ALL
      )

  def test_class_def(self):
    ret = execution.evaluate(
        inspect.cleandoc(
            """
            @dataclasses.dataclass
            class A:
              x: int
              y: int
              def __call__(self):
                return self.x + self.y
            """
        ),
        permission=permissions.CodePermission.ALL,
        global_vars=dict(dataclasses=dataclasses),
        outputs_intermediate=True,
    )
    self.assertEqual(list(ret.keys()), ['A', '__result__', '__stdout__'])
    self.assertTrue(inspect.isclass(ret['A']))
    self.assertIs(ret['__result__'], ret['A'])
    self.assertEqual(ret['__stdout__'], '')

  def test_function_def(self):
    ret = execution.evaluate(
        inspect.cleandoc(
            """
            def foo(x, y):
              return x + y

            def bar(z):
              return z + foo(z, z)
            """
        ),
        permission=permissions.CodePermission.ALL,
        outputs_intermediate=True,
    )
    self.assertEqual(
        list(ret.keys()), ['foo', 'bar', '__result__', '__stdout__']
    )
    self.assertTrue(inspect.isfunction(ret['foo']))
    self.assertTrue(inspect.isfunction(ret['bar']))
    self.assertIs(ret['__result__'], ret['bar'])

  def test_function_def_and_call(self):
    code = inspect.cleandoc(
        """
        def foo(x, y):
          return x + y

        def bar(z):
          print(f'z is {z}')
          return z + foo(z, z)

        bar(1)
        """
    )
    ret = execution.evaluate(
        code,
        permission=permissions.CodePermission.ALL,
        outputs_intermediate=True,
    )
    self.assertEqual(
        list(ret.keys()), ['foo', 'bar', '__result__', '__stdout__']
    )
    self.assertEqual(ret['__result__'], 3)
    ret = execution.evaluate(
        code,
        permission=permissions.CodePermission.ALL,
        returns_stdout=True,
    )
    self.assertEqual(ret, 'z is 1\n')

  def test_complex(self):
    ret = execution.evaluate(
        inspect.cleandoc(
            """
            @dataclasses.dataclass
            class A:
              x: int
              y: int
              def __call__(self, z):
                return self.x + self.y + z

            def foo(x, y):
              return x + y
            k = A(1, 2)
            k(foo(3, 4))
            """
        ),
        permission=permissions.CodePermission.ALL,
        global_vars=dict(dataclasses=dataclasses),
        outputs_intermediate=True,
    )
    self.assertEqual(
        list(ret.keys()), ['A', 'foo', 'k', '__result__', '__stdout__']
    )
    self.assertTrue(inspect.isclass(ret['A']))
    self.assertTrue(inspect.isfunction(ret['foo']))
    self.assertIsInstance(ret['k'], ret['A'])
    self.assertEqual(ret['__result__'], 10)

  def test_run_with_error(self):
    with self.assertRaisesRegex(
        errors.CodeError, 'NameError: name .* is not defined'
    ):
      execution.evaluate(
          inspect.cleandoc(
              """
              x = 1
              y = x + z
              """
          ),
          permission=permissions.CodePermission.ALL,
      )
    with self.assertRaisesRegex(errors.CodeError, 'ValueError'):
      execution.evaluate(
          'raise ValueError()', permission=permissions.CodePermission.ALL
      )


@dataclasses.dataclass
class Foo:
  x: int
  y: int


class SandboxCallTest(unittest.TestCase):

  def test_basics(self):
    def f(x, y):
      return x + y
    self.assertEqual(execution.sandbox_call(f, 1, y=2), 3)

  def test_complex_type(self):
    def f(x, y):
      return Foo(x, y)

    self.assertEqual(execution.sandbox_call(f, 1, 2), Foo(1, 2))

  def test_timeout(self):
    def f(x):
      time.sleep(x)

    self.assertIsNone(execution.sandbox_call(f, 0, timeout=1))
    with self.assertRaises(TimeoutError):
      execution.sandbox_call(f, 2, timeout=1)

  def test_raise(self):
    def f(x):
      if x == 0:
        raise ValueError()

    self.assertIsNone(execution.sandbox_call(f, 1))
    with self.assertRaises(ValueError):
      execution.sandbox_call(f, 0)


class CallTest(unittest.TestCase):

  def test_call_without_sandboxing(self):
    def foo(x, y):
      return x + y

    self.assertEqual(
        execution.maybe_sandbox_call(foo, 1, y=2, sandbox=False),
        3
    )

  def test_call_with_sandboxing(self):
    def foo(x, y):
      return x + y

    self.assertEqual(
        execution.maybe_sandbox_call(foo, 1, y=2, sandbox=True),
        3
    )

    def make_cls():
      @dataclasses.dataclass
      class A:
        x: str
      return A

    with self.assertRaises(errors.SerializationError):
      execution.maybe_sandbox_call(make_cls, sandbox=True)

  def test_call_with_automatic_sandboxing(self):
    def foo(x, y):
      return x + y

    self.assertEqual(
        execution.maybe_sandbox_call(foo, 1, y=2),
        3
    )

    def make_cls():
      @dataclasses.dataclass
      class A:
        x: str
      return A

    self.assertTrue(inspect.isclass(execution.maybe_sandbox_call(make_cls)))


class RunTest(unittest.TestCase):

  def test_run_without_sandboxing(self):
    self.assertEqual(
        execution.run(
            'x + y',
            global_vars=dict(x=1, y=2),
            sandbox=False,
        ),
        3,
    )

  def test_run_with_sandboxing(self):
    self.assertEqual(
        execution.run(
            'x + y',
            global_vars=dict(x=1, y=2),
            sandbox=True,
        ),
        3,
    )

  def test_run_with_automatic_sandboxing(self):
    self.assertEqual(
        execution.run(
            'x + y',
            global_vars=dict(x=1, y=2),
        ),
        3,
    )

    r = execution.run(
        inspect.cleandoc("""
            def foo(x, y):
              return x + y

            @dataclasses.dataclass
            class A:
              x: str
            """),
        global_vars=dict(dataclasses=dataclasses),
        outputs_intermediate=True,
    )
    self.assertTrue(inspect.isfunction(r['foo']))
    self.assertTrue(inspect.isclass(r['A']))


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import os
import unittest
from pyglove.core.utils import text_color


class TextColorTest(unittest.TestCase):

  def setUp(self):
    super().setUp()
    os.environ.pop('ANSI_COLORS_DISABLED', None)
    os.environ.pop('NO_COLOR', None)
    os.environ['FORCE_COLOR'] = '1'

  def test_colored_block_without_colors_and_styles(self):
    self.assertEqual(text_color.colored_block('foo', '{{', '}}'), 'foo')

  def test_colored_block(self):
    original_text = inspect.cleandoc("""
        Hi << foo >>
        <# print x if x is present #>
        <% if x %>
        << x >>
        <% endif %>
        """)

    colored_text = text_color.colored_block(
        text_color.colored(original_text, color='blue'),
        '<<', '>>',
        color='white',
        background='blue',
    )
    origin_color = '\x1b[34m'
    reset = '\x1b[0m'
    block_color = text_color.colored(
        'TEXT', color='white', background='blue'
    ).split('TEXT')[0]
    self.assertEqual(
        colored_text,
        f'{origin_color}Hi {block_color}<< foo >>{reset}{origin_color}\n'
        '<# print x if x is present #>\n<% if x %>\n'
        f'{block_color}<< x >>{reset}{origin_color}\n'
        f'<% endif %>{reset}'
    )
    self.assertEqual(text_color.decolor(colored_text), original_text)

  def test_colored_block_without_full_match(self):
    self.assertEqual(
        text_color.colored_block(
            'Hi {{ foo',
            '{{', '}}',
            color='white',
            background='blue',
        ),
        'Hi {{ foo'
    )

  def test_colored_block_without_termcolor(self):
    termcolor = text_color.termcolor
    text_color.termcolor = None
    original_text = inspect.cleandoc("""
        Hi {{ foo }}
        {# print x if x is present #}
        {% if x %}
        {{ x }}
        {% endif %}
        """)

    colored_text = text_color.colored_block(
        text_color.colored(original_text, color='blue'),
        '{{', '}}',
        color='white',
        background='blue',
    )
    self.assertEqual(colored_text, original_text)
    self.assertEqual(text_color.decolor(colored_text), original_text)
    text_color.termcolor = termcolor


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Conversion from annotations to PyGlove value specs."""

import collections
import inspect
import types
import typing

from pyglove.core import utils
from pyglove.core.typing import annotated
from pyglove.core.typing import class_schema
from pyglove.core.typing import inspect as pg_inspect
from pyglove.core.typing import key_specs as ks
from pyglove.core.typing import value_specs as vs


_NoneType = type(None)

# Annotated is suppored after 3.9
_Annotated = getattr(typing, 'Annotated', None)  # pylint: disable=invalid-name

# UnionType is supported after 3.10.
_UnionType = getattr(types, 'UnionType', None)  # pylint: disable=invalid-name


def _field_from_annotation(
    key: typing.Union[str, class_schema.KeySpec],
    annotation: typing.Any,
    description: typing.Optional[str] = None,
    metadata: typing.Optional[typing.Dict[str, typing.Any]] = None,
    auto_typing: bool = True,
    parent_module: typing.Optional[types.ModuleType] = None
) -> class_schema.Field:
  """Creates a field from Python annotation."""
  if isinstance(annotation, annotated.Annotated):
    field_spec = (
        key, annotation.value_spec, annotation.docstring, annotation.metadata)
  elif _Annotated and typing.get_origin(annotation) is _Annotated:
    type_args = typing.get_args(annotation)
    assert len(type_args) > 1, (annotation, type_args)
    field_spec = tuple([key] + list(type_args))
  else:
    field_spec = (key, annotation, description, metadata or {})
  return class_schema.create_field(
      field_spec,
      auto_typing=auto_typing,
      accept_value_as_annotation=False,
      parent_module=parent_module
  )


def _value_spec_from_default_value(
    value: typing.Any,
    set_default: bool = True) -> typing.Optional[class_schema.ValueSpec]:
  """Creates a value spec from a default value."""
  if isinstance(value, (bool, int, float, str, dict)):
    value_spec = _value_spec_from_type_annotation(type(value), False)
  elif isinstance(value, list):
    value_spec = vs.List(
        _value_spec_from_default_value(value[0], False) if value else vs.Any())
  elif isinstance(value, tuple):
    value_spec = vs.Tuple(
        [_value_spec_from_default_value(elem, False) for elem in value])
  elif inspect.isfunction(value) or isinstance(value, utils.Functor):
    value_spec = vs.Callable()
  elif not isinstance(value, type):
    value_spec = vs.Object(type(value))
  else:
    value_spec = None

  if value_spec and set_default:
    value_spec.set_default(value)
  return value_spec


def _value_spec_from_type_annotation(
    annotation: typing.Any,
    accept_value_as_annotation: bool,
    parent_module: typing.Optional[types.ModuleType] = None
) -> class_schema.ValueSpec:
  """Creates a value spec from type annotation."""
  if annotation is bool:
    return vs.Bool()
  elif annotation is int:
    return vs.Int()
  elif annotation is float:
    return vs.Float()
  elif annotation is str:
    return vs.Str()
  elif annotation is typing.Any or annotation is vs.Any:
    return vs.Any().annotate(annotation)

  origin = typing.get_origin(annotation) or annotation
  args = list(typing.get_args(annotation))

  def _sub_value_spec_from_annotation(
      annotation: typing.Any) -> class_schema.ValueSpec:
    return _value_spec_from_type_annotation(
        annotation, accept_value_as_annotation, parent_module)

  # Handling list.
  if origin in (list, typing.List):
    return vs.List(
        _sub_value_spec_from_annotation(args[0])) if args else vs.List(vs.Any())
  # Handling tuple.
  elif origin in (tuple, typing.Tuple):
    if not args:
      return vs.Tuple(vs.Any())
    else:
      if args[-1] is ...:
        if len(args) != 2:
          raise TypeError(
              f'Tuple with ellipsis should have exact 2 type arguments. '
              f'Encountered: {annotation}.')
        return vs.Tuple(_sub_value_spec_from_annotation(args[0]))
      return vs.Tuple([_sub_value_spec_from_annotation(arg) for arg in args])
  # Handling sequence.
  elif origin in (collections.abc.Sequence,):
    elem = _sub_value_spec_from_annotation(args[0]) if args else vs.Any()
    return vs.Union([vs.List(elem), vs.Tuple(elem)])
  # Handling literals.
  elif origin is typing.Literal:
    return vs.Enum(utils.MISSING_VALUE, args)
  # Handling dict.
  elif origin in (dict, typing.Dict, collections.abc.Mapping):
    if not args:
      return vs.Dict()
    assert len(args) == 2, (annotation, args)
    if args[0] not in (str, typing.Text):
      raise TypeError(
          'Dict type field with non-string key is not supported.')
    elem_value_spec = _sub_value_spec_from_annotation(args[1])
    return vs.Dict([(ks.StrKey(), elem_value_spec)])
  elif origin is collections.abc.Callable:
    arg_specs = []
    return_spec = None
    if args:
      assert len(args) == 2, (annotation, args)

      # NOTE: Various ways of expressing the input args of `Callable` will
      # be normalized into a list (see examples below). Therefore, we only
      # check against list here.
      #
      #   Callable[int, Any] => Callable[[int], Any]
      #   Callable[(int, int), Any] => Callable[[int, int], Any]
      if isinstance(args[0], list):
        arg_specs = [_sub_value_spec_from_annotation(arg) for arg in args[0]]
      return_spec = _sub_value_spec_from_annotation(args[1])
    return vs.Callable(arg_specs, returns=return_spec)
  # Handling type
  elif origin is type or (annotation in (typing.Type, type)):
    if not args:
      return vs.Type(typing.Any)
    assert len(args) == 1, (annotation, args)
    return vs.Type(args[0])
  # Handling union.
  elif origin is typing.Union or (_UnionType and origin is _UnionType):
    optional = _NoneType in args
    if optional:
      args.remove(_NoneType)
    if len(args) == 1:
      spec = _sub_value_spec_from_annotation(args[0])
    else:
      spec = vs.Union([_sub_value_spec_from_annotation(x) for x in args])
    if optional:
      spec = spec.noneable()
    return spec
  elif origin is typing.Final:
    return _value_spec_from_type_annotation(
        args[0],
        accept_value_as_annotation=False
    ).freeze(vs._FROZEN_VALUE_PLACEHOLDER)  # pylint: disable=protected-access
  elif isinstance(annotation, typing.ForwardRef):
    annotation = annotation.__forward_arg__
    if parent_module is not None:
      annotation = class_schema.ForwardRef(parent_module, annotation)
    return vs.Object(annotation)
  elif isinstance(annotation, class_schema.ForwardRef):
    return vs.Object(annotation)
  # Handling class.
  elif (
      inspect.isclass(annotation)
      or pg_inspect.is_generic(annotation)
      or (isinstance(annotation, str) and not accept_value_as_annotation)
  ):
    if isinstance(annotation, str) and parent_module is not None:
      annotation = class_schema.ForwardRef(parent_module, annotation)
    return vs.Object(annotation)

  if accept_value_as_annotation:
    spec = _value_spec_from_default_value(annotation)
    if spec is not None:
      return spec
  raise TypeError(
      f'Cannot convert {annotation!r} to `pg.typing.ValueSpec` '
      f'with auto typing.')


def _any_spec_with_annotation(annotation: typing.Any) -> vs.Any:
  """Creates an ``Any`` value spec with annotation."""
  value_spec = vs.Any()
  if annotation != inspect.Parameter.empty:
    value_spec.annotate(annotation)
  return value_spec


def _value_spec_from_annotation(
    annotation: typing.Any,
    auto_typing: bool = False,
    accept_value_as_annotation: bool = False,
    parent_module: typing.Optional[types.ModuleType] = None
    ) -> class_schema.ValueSpec:
  """Creates a value spec from annotation."""
  if isinstance(annotation, class_schema.ValueSpec):
    return annotation
  elif annotation == inspect.Parameter.empty:
    return vs.Any()
  elif annotation is None:
    if accept_value_as_annotation:
      return vs.Any().noneable()
    else:
      return vs.Any().freeze(None)

  if auto_typing:
    return _value_spec_from_type_annotation(
        annotation, accept_value_as_annotation, parent_module
    )
  else:
    value_spec = None
    if accept_value_as_annotation:
      # Accept default values is applicable only when auto typing is off.
      value_spec = _value_spec_from_default_value(annotation)
    return value_spec or _any_spec_with_annotation(annotation)


class_schema.Field.from_annotation = _field_from_annotation
class_schema.ValueSpec.from_annotation = _value_spec_from_annotation
